require "defines"
require "util"

--[[ 
Everything will be prefixed by ldLC_
--]]

-----------------------------------------------------------
-- utility functions
-- /c print = function(s)  game.player.print(s) end
-- /c sel = function() return game.player.selected end
-- /c print("help = " .. sel().help())
-----------------------------------------------------------

-----------------------------------------------------------
-- utility functions
-----------------------------------------------------------
local function ldLC_Debug(s)
  for index, player in pairs(game.players) do
    if(player.valid) then
      player.print(s);
    end;
  end
end

-----------------------------------------------------------
-- utility functions
-----------------------------------------------------------
local function ldLC_DebugObj(prefix, o)
  if (o == nil)
  then
    ldLC_Debug(prefix .. " is nil")
  else
    ldLC_Debug(prefix .. " name=" .. o.name .. ", type=" .. o.type)
  end
end

-----------------------------------------------------------
-- utility functions
-- return the current player object.
-----------------------------------------------------------
local function ldLC_getPlayer(event)
  -- multiplayer
  --[[
  if(nil ~= event and nil ~= event.player_index) then
    ldLC_Debug("event.player_index = " .. event.player_index);
  else
    ldLC_Debug("event.player_index = nil");
  end
  --]]
  if(event ~= nil and event.player_index ~= nil and event.player_index > 0 and game.players[event.player_index] ~= nil)
  then
    return game.players[event.player_index];
  end;
  if(#game.players > 1)
  then
    for index, player in pairs(game.players) do
      if(player.valid) then
        return player;
      end;
    end
  end
  return game.player;
end;

-----------------------------------------------------------
-- utility functions
-----------------------------------------------------------
local function ldLC_IsLargeChest(name)
  if    (name == "ldLC-chest-2x2-obj") 
     or (name == "ldLC-chest-3x3-obj") 
     or (name == "ldLC-chest-5x5-obj") 
  then
    return true
  else
    return false
  end
end

-----------------------------------------------------------
-- when a chest is created
-----------------------------------------------------------
function ldLC_CreateLC(event)
  if ldLC_IsLargeChest(event.created_entity.name) 
  then
    local chest = event.created_entity
    local surface = chest.surface;
    local wagon_name = string.gsub(chest.name, "-obj$", "")
    local position = chest.position
    -- ldLC_DebugObj("chest -- ", chest)
    chest.destroy()
    local rail = surface.create_entity{name="ldLC-temporary-rail", position = position, direction = 0,force=game.forces.player}
    -- ldLC_DebugObj("rail  -- ", rail)
    local wagon = surface.create_entity{name=wagon_name, position = position, direction = 0,force=game.forces.player}
    -- ldLC_DebugObj("wagon -- ", wagon)
    -- rail.destroy()
    event.created_entity = wagon
    -- for inv in chest.getinventory(1) do
    --   inv.setfilter("coin")
    -- end
    -- chest.setfilter(1, "coin")
  end
end

-----------------------------------------------------------
-- when a chest is mined
-----------------------------------------------------------
local function ldLC_IsLoneTempRail(rail)
  if (rail ~= nil and rail.valid and rail.name == "ldLC-temporary-rail" and rail.minable)
  then
    return true
  else
    return false
  end
end

-----------------------------------------------------------
-- when a chest is mined
-----------------------------------------------------------
function ldLC_MinedLC(event)
  if ldLC_IsLargeChest(event.item_stack.name) 
  then
    local player = ldLC_getPlayer(event);
    local rail = player.selected;
    local surface = player.surface;
    -- ldLC_Debug("pidx  -- " .. event.player_index .. " -- name = " .. player.name);
    -- ldLC_DebugObj("rail  -- ", rail)
    if (not ldLC_IsLoneTempRail(rail))
    then
      -- see if we can find the piece of rail that is left befind...
      -- local cursor = player.screen2realposition(player.cursorposition);
      local cursor = player.position;
      -- ldLC_Debug("cursor-- ", cursor.x .. "," .. cursor.y)
      local player_reach = 8;
      local width = 4 + player_reach
      local height = 4 + player_reach
      if (event.item_stack.name == "ldLC-chest-5x5-obj")
      then
        width = 6 + player_reach
        height = 6 + player_reach
      end
      -- ldLC_Debug("area  -- { {" .. (cursor.x - width) .. "," .. (cursor.y - height) .. "}, {" .. (cursor.x + width) .. ", " .. (cursor.y + height) .. "} }")
      local entities = surface.find_entities_filtered{area = {{cursor.x - width, cursor.y - height}, {cursor.x + width, cursor.y + height}}, name = "ldLC-temporary-rail"}
      for key,value in pairs(entities)
      do
        -- ldLC_DebugObj("value -- ", value)
        if (ldLC_IsLoneTempRail(value))
        then
          rail = value
          break
        end
      end
    end
    if (ldLC_IsLoneTempRail(rail))
    then
      rail.destroy()
    end
  end
end

-----------------------------------------------------------
-- register our events
-----------------------------------------------------------
game.on_event(defines.events.on_built_entity, ldLC_CreateLC)
game.on_event(defines.events.on_player_mined_item, ldLC_MinedLC)
-- since our "chest" is a wagon on rail, the blueprints will only capture the rail.
-- as a result, robots can't mine it.
game.on_event(defines.events.on_robot_built_entity, ldLC_CreateLC)
game.on_event(defines.events.on_robot_mined, ldLC_MinedLC)
